"""
Functions for interacting with NCBI databases and file types.
"""

from topiary import util, check

import pandas as pd
import numpy as np

# Modules for blasting, etc.
from Bio import SeqIO, Entrez
from Bio.Seq import Seq
from Bio.Blast import NCBIXML, NCBIWWW
import Bio.Blast.Applications as apps
from Bio import pairwise2

from tqdm.auto import tqdm

import re, sys, os, string, random, pickle, io, urllib, http, subprocess
import multiprocessing as mp

def _grab_line_meta_data(line):
    """
    Look for meta data we care about from the line.  This includes:

        + structure
        + low_quality
        + precursor
        + predicted
        + isoform
        + hypothetical
        + partial

    Return a dictionary with each of those keyed to a bool for whether
    or not the line has this.

    Parameters
    ----------
        line: line from an NCBI record of some sort

    Return
    ------
        dictionary with true or false for different patterns in line
    """

    meta_patterns = {"structure":"crystal structure",
                     "low_quality":"low.quality",
                     "predicted":"predicted",
                     "precursor":"precursor",
                     "isoform":"isoform",
                     "hypothetical":"hypothetical",
                     "partial":"partial"}

    out = dict([(p,None) for p in meta_patterns])
    for m in meta_patterns:
        mp = re.compile(meta_patterns[m],re.IGNORECASE)
        out[m] = bool(mp.search(line))

    return out


def parse_ncbi_line(line,accession=None):
    """
    Parse an ncbi line of the sort seen in the BLAST title field or on each
    line of a fasta file.

    Parameters
    ----------
        line: line from an NCBI record
        accession: extract entry from line that matches acccession.  Ignores
                   version (e.g. "1" in XXXXXXXXX.1).  If None, parse first
                   entry on line.

    Return
    ------
        dictionary with following keys:
            raw_line: unprocessed line (input)
            line: processed line (remove multiple titles)
            name: protein name
            structure: whether or not this is a structure (bool)
            low_quality: whether or not this is low quality (bool)
            predicted: whether or not this is predicted (bool)
            precursor: whether or not this is a precursor (bool)
            isoform: whether or not this is an isoform (bool)
            hypothetical: whether or not this is hypothetical (bool)
            partial: whether or not this is a partial sequence (bool)
    """

    out = {"raw_line":line}

    # Clean up leading and trailing spaces
    line = line.strip()

    # Clean up leading ">" if it's there (difference between fasta and blast
    # lines)
    if line.startswith(">"):
        line = line[1:]

    # Particularly in BLAST hits, there are often multiple entries separated
    # by ">" (basically, pasted together fasta entries). Only take one.  If
    # accession is specified, take that accesion.  If no accession is
    # specified, take the first one.

    # Split on ">"
    line_splitter = re.compile(">")
    entries_on_line = [s.strip() for s in line_splitter.split(line)]

    # Extract accession from entry (assumes xxx|acccession stuff).  Ignores
    # trailing accession version number XXXXXXXX.1 -> XXXXXXXX
    accession_dict = {}
    for e in entries_on_line:
        try:
            if e.startswith("pdb"):
                k = [v.split()[0] for v in e.split("|")[1:3]]
                k = "{}_{}".format(*k)
            else:
                k = e.split("|")[-2].split(".")[0].strip()
            accession_dict[k] = e
        except IndexError:
            print(f"Could not parse line {e}. Skipping.")
            return None

    # If accession is specified, grab entry corresponding to that from the
    # line.  If not, grab first entry.
    if not accession:
        accession = list(accession_dict.keys())[0]

    accession = accession.split(".")[0].strip()
    try:
        line = accession_dict[accession]
    except KeyError:
        warn = f"accession '{accession}' not found in line '{line}'\n"
        warn += f"Using accession {accession} anyway and returning entire line."
        print(warn)

    out["accession"] = accession

    # Record the processed line
    out["line"] = line

    # Grab meta stuff (predicted, isoform, etc.)
    meta = _grab_line_meta_data(line)
    for m in meta:
        out[m] = meta[m]

    # Look for species name (thing within [ xxx ])
    species = None
    species_pattern = re.compile("\[.*?]")

    sm = species_pattern.search(line)
    if sm:
        out["species"] = sm.group(0)[1:-1]
    else:
        out["species"] = None

    # Clean up any double spaces introduced into the line at this point
    line = re.sub("  "," ",line)

    # Protein name (takes between '| XXXXXXXXX [' ).
    out["name"] = line.split("|")[-1].split("[")[0].strip()

    return out



"""
Pipeline that starts from an alignment dataframe ultimately creates ancestors.
"""

import topiary

def generate_ancestors(df,output_dir,do_bootstrap=True):
    """
    """

    topiary.find_best_model(df,
                            output="00_find-model")

    topiary.generate_ml_tree(previous_dir="00_find-model",
                             output="01_ml-tree",
                             bootstrap=do_bootstrap)

    topiary.reconcile(previous_dir="01_ml-tree",
                      output="02_reconciliation")

    topiary.generate_ancestors(previous_dir="02_reconciliation",
                               output="03_ancestors")



"""
Input/output functions for topiary.
"""

from .dataframe import read_dataframe, write_dataframe
from .construct import df_from_blast_xml, df_from_seed
from .alignments import write_fasta, write_phy, read_fasta_into
from .tree import read_tree
from .seed import load_seed_dataframe



"""
Interface to NCBI blast and entrez.
"""

from Bio import Entrez
Entrez.email = "topiary.phylogenetics@gmail.com"

from . _parse_ncbi_line import parse_ncbi_line
from . blast import read_blast_xml, local_blast, merge_blast_df, ncbi_blast, recip_blast, make_blast_db
from . entrez import get_sequences, get_taxid, get_proteome



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import io
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command, Extension

# Package meta-data.
NAME = 'topiary-asr'
DESCRIPTION = "Python framework for doing ancestral sequence reconstruction using pandas dataframes and ete3 trees as the primary data structures."
URL = 'https://github.com/harmslab/topiary'
EMAIL = 'harmsm@gmail.com'
AUTHOR = 'Michael J. Harms'
REQUIRES_PYTHON = '>=3.8.0'
VERSION = None

# What packages are required for this module to be executed?
REQUIRED = [

    # General scientific computing and display
    "numpy",
    "pandas>=1.3.1",
    "xlrd",
    "openpyxl",
    "matplotlib>=3.4.2",
    "tqdm",
    "pyqt5",

    # bioinformatics
    "biopython>=1.78",
    "ete3>=3.1.2",
    "opentree>=1.0.1",
    "dendropy>=4.5.2",
    "pastml>=1.9.34"
]


here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = '\n' + f.read()

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    with open(os.path.join(here, "topiary", '__version__.py')) as f:
        exec(f.read(), about)
else:
    about['__version__'] = VERSION

# Where the magic happens:
setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=('tests',)),
    install_requires=REQUIRED,
    extras_require = {
        'test': ['pytest'],
    },
    scripts=['bin/run-raxml','bin/setup-generax'],
    package_data={'': ['*.csv']},
    include_package_data=True,
    license='MIT',
    classifiers=[
      'Development Status :: 3 - Alpha',
      'Intended Audience :: Science/Research',
      'Programming Language :: Python :: 3.8',
      'Programming Language :: Python :: 3.9',
    ],
    keywords='phylogenetics ASR'
)



"""
Run BLAST against a local database.
"""

import topiary
from topiary import check
from .util import read_blast_xml, _standard_blast_args_checker

import Bio.Blast.Applications as apps

import numpy as np
import pandas as pd
from tqdm.auto import tqdm

import sys, os, string, random, subprocess, copy
import multiprocessing as mp

def _prepare_for_blast(sequence,
                       db,
                       blast_program,
                       hitlist_size,
                       e_value_cutoff,
                       gapcosts,
                       kwargs,
                       test_skip_blast_program_check=False):
    """
    Take inputs to local_blast, check arguments, and do initial processing.
    Outputs are a list of validated sequences, the blast function to use,
    a dictionary of keyword arguments to past to blast for each query, and
    whether or not to return a single df or list of df at the end.

    Parameters
    ----------
        sequence: sequence as a string OR list of string sequences
        db: name of local blast database
        blast_program: NCBI blast program to use (blastp, tblastn, etc.)
        hitlist_size: download only the top hitlist_size hits
        e_value_cutoff: only take hits with e_value better than e_value_cutoff
        gapcosts: BLAST gapcosts (length 2 tuple of ints)
        kwargs: extra keyword arguments are passed directly to
                apps.NcbiblastXXXCommandline, overriding anything constructed
                above.
        test_skip_blast_program_check: skip the check for a working blast
                                       program (for testing)

    Return
    ------
        dataframe (single query) or list of dataframes (multiple sequences).
        if no hits found, warns and returns None.
    """

    recognized_functions = {"blastp":apps.NcbiblastpCommandline,
                            "blastn":apps.NcbiblastnCommandline,
                            "blastx":apps.NcbiblastxCommandline,
                            "tblastn":apps.NcbitblastnCommandline,
                            "tblastx":apps.NcbitblastxCommandline,
                            "psiblast":apps.NcbipsiblastCommandline,
                            "rpsblast":apps.NcbirpsblastCommandline,
                            "rpstblastn":apps.NcbirpstblastnCommandline,
                            "deltablast":apps.NcbideltablastCommandline}

    if not os.path.exists(f"{db}.psq"):
        err = f"db {db}.psq not found!\n"
        raise FileNotFoundError(err)

    # Make sure we can actually run the local blasting
    try:
        blast_function = recognized_functions[blast_program]
    except (KeyError,TypeError):
        err = "\nblast_program '{}' not recognized.\n\nAllowed programs:\n".format(blast_program)
        for k in recognized_functions.keys():
            err += "    {}\n".format(k)
        raise ValueError(err)

    # Make sure the blast program is installed.
    if not test_skip_blast_program_check:
        try:
            ret = subprocess.run([blast_program],stderr=subprocess.DEVNULL)
        except FileNotFoundError:
            err = f"\nBLAST program {blast_program} is not in path. Is it installed?\n\n"
            raise FileNotFoundError(err)

    out = _standard_blast_args_checker(sequence,
                                       hitlist_size,
                                       e_value_cutoff,
                                       gapcosts)
    sequence_list = out[0]
    hitlist_size = out[1]
    e_value_cutoff = out[2]
    gapcosts = out[3]
    return_singleton = out[4]

    gaps = '{} {}'.format(*gapcosts)

    # Construct keyword arguments to pass to function
    blast_kwargs = {"cmd":blast_program,
                    "db":db,
                    "outfmt":5,
                    "max_target_seqs":hitlist_size,
                    "threshold":e_value_cutoff,
                    "gapopen":gapcosts[0],
                    "gapextend":gapcosts[1]}
    for k in kwargs:
        blast_kwargs[k] = kwargs[k]


    return sequence_list, blast_function, blast_kwargs, return_singleton

def _construct_args(sequence_list,
                    blast_function,
                    blast_kwargs,
                    keep_tmp=False,
                    block_size=20,
                    num_threads=-1,
                    test_num_cores=None):
    """
    Construct a list of arguments to pass to each thread in the pool.

    Parameters
    ----------
        sequence_list: list of sequences as strings
        blast_function: blast function to use
        blast_kwargs: keyword arguments to pass to blast call
        keep_tmp: whether or not to keep temporary files
        num_threads: number of threads to use. if -1, use all available.
        block_size: break into block_size sequence chunks

    Return
    ------
        list of args to pass for each calculation, number of threads
    """

    # Validate inputs that have not yet been validated.
    block_size = check.check_int(block_size,
                                             "block_size",
                                              minimum_allowed=1)

    num_threads = check.check_int(num_threads,
                                              "num_threads",
                                              minimum_allowed=-1)
    if num_threads == 0:
        err = "\nnum_threads cannot be zero. It can be -1 (use all available),\n"
        err += "or any integer > 0.\n\n"
        raise ValueError(err)

    keep_tmp = check.check_bool(keep_tmp,"keep_tmp")

    # Try to figure out how many cores are available
    if test_num_cores is not None:
        num_cores = test_num_cores
    else:
        try:
            num_cores = mp.cpu_count()
        except NotImplementedError:
            num_cores = os.cpu_count()

        # If we can't figure it out, revert to 1
        if num_cores is None:
            print("Could not determine number of cpus. Using single thread.",flush=True)
            num_cores = 1

    # Determine number of threads useful for this problem. It's not worth
    # chopping up a super small set of comparisons
    max_useful_threads = len(sequence_list)//block_size
    if max_useful_threads < 1:
        max_useful_threads = 1
    if max_useful_threads > num_cores:
        max_useful_threads = num_cores

    # Set number of threads
    if num_threads == -1 or num_threads > max_useful_threads:
        num_threads = max_useful_threads

    # Break sequences up into blocks
    num_sequences = len(sequence_list)

    num_windows = num_sequences//block_size
    if num_windows == 0:
        num_windows = 1
        block_size = num_sequences

    windows = [block_size for _ in range(num_windows)]
    remainder = num_sequences % block_size
    if remainder/block_size >= 0.5:
        windows.append(remainder)
    else:
        counter = 0
        while remainder > 0:
            windows[counter] += 1
            remainder -= 1
            counter += 1

    windows.insert(0,0)
    windows = np.cumsum(windows)

    # Blocks will allow us to tile over whole sequence
    all_args = []
    for i in range(len(windows)-1):

        i_block = (windows[i],windows[i+1])

        all_args.append([sequence_list,
                         i_block,
                         blast_function,
                         blast_kwargs,
                         keep_tmp])

    return all_args, num_threads


def _thread_manager(all_args,num_threads):
    """
    Run a bunch of blast jobs in a mulithreaded fashion. Should only be called
    by local_blast.

    Parameters
    ----------
        all_args: list of args to pass for each calculatio
        nm_threads: number of threads to use

    Return
    ------
        list of dataframes with blast results
    """

    print(f"Performing {len(all_args)} blast queries on {num_threads} threads.",
          flush=True)

    # queue will hold results from each run. Append to each entry in all_args
    queue = mp.Manager().Queue()
    for i in range(len(all_args)):
        all_args[i].append(queue)

    with mp.Pool(num_threads) as pool:

        # Black magic. pool.imap() runs a function on elements in iterable,
        # filling threads as each job finishes. (Calls _blast_thread
        # on every args tuple in all_args). tqdm gives us a status bar.
        # By wrapping pool.imap iterator in tqdm, we get a status bar that
        # updates as each thread finishes.
        list(tqdm(pool.imap(_thread,all_args),total=len(all_args)))

    # Get results out of the queue.
    results = []
    while not queue.empty():
        results.append(queue.get())

    # Sort results
    results.sort()
    hits = [r[1] for r in results]

    return hits

def _thread(args):
    """
    Run blast on a thread. Should only be called via _thread_manager.
    Puts resulting hits as a pandas dataframe into the queue

    Parameters
    ----------
        args. list that is expanded as follows:

        sequence_list: list of sequences as strings
        index: sequence to grab
        blast_kwargs: keyword arguments to pass to blast call
        blast_function: blast function to call
        keep_tmp: whether or not to keep temporary files
        queue: multiprocessing queue for storing results

    Return
    ------
        None
    """

    # parse args
    sequence_list = args[0]
    index = args[1]
    blast_function = args[2]
    blast_kwargs = args[3]
    keep_tmp = args[4]
    queue = args[5]

    # make a 10-character random string for temporary files
    tmp_file_root = "".join([random.choice(string.ascii_letters) for i in range(10)])
    input_file = "topiary-tmp_{}_blast-in.fasta".format(tmp_file_root)
    out_file = "topiary-tmp_{}_blast-out.xml".format(tmp_file_root)

    f = open(input_file,'w')
    for i in range(index[0],index[1]):
        f.write("".join(f">count{i}\n{sequence_list[i]}\n"))
    f.close()

    blast_kwargs = copy.deepcopy(blast_kwargs)
    blast_kwargs["query"] = input_file
    blast_kwargs["out"] = out_file

    blast_function(**blast_kwargs)()

    # Parse output
    try:
        out_df = read_blast_xml(out_file)
    except FileNotFoundError:
        err = "\nLocal blast failed on sequence:\n"
        err += f"    '{sequence_list[i]}'\n\n"
        raise RuntimeError(err)

    # Delete temporary files
    if not keep_tmp:
        os.remove(input_file)
        os.remove(out_file)

    # update queue
    queue.put((index[0],out_df))

def _combine_hits(hits,return_singleton):
    """
    Parse a list of hits resulting from a set of blast queries and return a
    list of dataframes, one for each query. Results will be ordered first by the
    order they occur in the list, then sorted by query. (Assumes queries have
    the form countNUMBER when it does sorting.

    Parameters
    ----------
        hits: list of dataframes containing hits
        return_singleton: bool. whether or not to return a single dataframe or
                          list of dataframes

    Return
    ------
        combined blast hits (list of dataframes or single dataframe)
    """

    # Construct a list of output dataframes, one for each query sequence
    out_df = []
    for h in hits:
        queries = np.unique(h["query"])
        for q in queries:
            this_df = h.loc[h["query"] == q,:]

            # No hits, return empty dataframe
            if len(this_df) == 1 and pd.isna(this_df["accession"].iloc[0]):
                out_df.append(pd.DataFrame())

            # Record hits
            else:
                out_df.append(this_df)

    # Singleton hit -- pop out of list
    if return_singleton:
        out_df = out_df[0]

    return out_df


def local_blast(sequence,
                db,
                blast_program="blastp",
                hitlist_size=100,
                e_value_cutoff=0.001,
                gapcosts=(11,1),
                keep_tmp=False,
                num_threads=-1,
                block_size=20,
                **kwargs):
    """
    Perform a blast query against a local blast database. Takes a sequence or
    list of sequences and returns a list of topiary dataframes containing hits
    for each sequence.

    Parameters
    ----------
    sequence : str or list
        sequence as a string OR list of string sequences
    db : str
        name of local blast database
    blast_program : str, default="blastp"
        NCBI blast program to use (i.e. "blastp", "tblastn", etc.)
    hitlist_size : int, default=100
        return only the top hitlist_size hits
    e_value_cutoff : float, default=0.001
        only return hits with e_value better than e_value_cutoff
    gapcosts : tuple, default=(11,1)
        BLAST gapcosts (length 2 tuple of ints)
    keep_tmp : bool, default=False
        whether or not to keep temporary blast output
    num_threads : int, default=-1
        number of threads to use. if -1, use all available.
    block_size : int, default=20
        run blast in blocks of block_size sequences
    **kwargs : dict, optional
        extra keyword arguments are passed directly to
        apps.NcbiblastXXXCommandline.

    Returns
    -------
    blast_output : pandas.DataFrame or list
        If a single sequence is passed in, return a dataframe of hits. If a list
        of queries is passed in, returns a list of dataframes--one entry per
        query. If no hits are found for a given query, return an empty dataframe.
    """

    # Generate BLAST input
    prep = _prepare_for_blast(sequence=sequence,
                              db=db,
                              blast_program=blast_program,
                              hitlist_size=hitlist_size,
                              e_value_cutoff=e_value_cutoff,
                              gapcosts=gapcosts,
                              kwargs=kwargs)

    sequence_list = prep[0]
    blast_function = prep[1]
    blast_kwargs = prep[2]
    return_singleton = prep[3]

    # Construct a list of arguments to pass into _thread_manager
    all_args, num_threads = _construct_args(sequence_list=sequence_list,
                                            blast_function=blast_function,
                                            blast_kwargs=blast_kwargs,
                                            keep_tmp=keep_tmp,
                                            block_size=block_size,
                                            num_threads=num_threads)

    # Run multi-threaded local blast
    hits = _thread_manager(all_args,num_threads)

    # Combine hits into dataframes, one for each query
    out_df = _combine_hits(hits,return_singleton)

    return out_df



"""
Use entrez to get the NCBI taxid for species.
"""

import topiary
from topiary import check

from Bio import Entrez

def get_taxid(species_list):
    """
    Use entrez to get the NCBI taxid for species.

    Parameters
    ----------
    species_list : list
        list of species in binomial format (i.e. Homo sapiens).

    Returns
    -------
    taxid_list : list
        list of taxid (*not* guaranteed to be in the same order as the input
        species_list)
    """

    # Make sure species list is sane, each species is unique, and the list
    # is sorted
    species_list = check.check_iter(species_list,
                                                "species_list",
                                                required_value_type=str)
    return_singleton = False
    if type(species_list) is str:
        species_list = [species_list]
        return_singleton = True

    species_list = list(set(species_list))
    species_list.sort()

    # return nothing if nothing was passed in
    if len(species_list) == 0:
        return []

    # Create a search term
    search_term = " OR ".join(species_list)

    # Access Entrez and download
    handle = Entrez.esearch(db="taxonomy",
                            retmax=len(species_list)*2,
                            term=search_term,
                            idtype="uilist")
    record = Entrez.read(handle)

    # See how many records we pulled down
    count = int(record["Count"])

    if count != len(species_list):

        error_list = []

        error = record["ErrorList"]
        for k in error:
            if len(error[k]) > 0:
                error_list.append(f"{k}: {error[k]}")

        err = "\nEntrez returned the following error when retrieving taxids. This\n"
        err += "caused by mis-spelled species name(s).\n\n"
        for e in error_list:
            err += f" -> {e}\n"
        err += "\n"

        raise RuntimeError(err)

    taxid_list = [str(i) for i in record["IdList"]]

    if return_singleton:
        taxid_list = taxid_list[0]

    return taxid_list



"""
Shared functions for dealing with NCBI blast.
"""

from topiary import util, check

import pandas as pd
import numpy as np

# Modules for blasting, etc.
from Bio.Blast import NCBIXML

def _standard_blast_args_checker(sequence,
                                 hitlist_size,
                                 e_value_cutoff,
                                 gapcosts):
    """
    Parse and validate input used by all blast queries.

    Parameters
    ----------
        sequence: sequence as a string OR list of string sequences
        hitlist_size: download only the top hitlist_size hits
        e_value_cutoff: only take hits with e_value better than e_value_cutoff
        gapcosts: BLAST gapcosts (length 2 tuple of ints)

    Return
    ------
        list of sequence, hitlist_size, e_value_cutoff, gapcosts (tuple), and
        whether or not to return final output as a single sequence or list of
        sequences.
    """

    # Parse the sequence input. Should either be a list of strings or a single
    # string sequence. Create sequence_list which is a list of sequences (even
    # if only one)
    return_singleton = False
    if hasattr(sequence,"__iter__") and type(sequence) is not type:

        # If sequence is not a string, make a single fasta-formatted string
        # from contents
        sequence_list = []
        if type(sequence) is not str:
            for i, s in enumerate(sequence):
                if type(s) is not str:
                    err = "\nsequence must either be a string or list of strings, one for each sequence\n\n"
                    raise ValueError(err)

                seq = s.strip()
                if len(seq) == 0:
                    err = "\nempty sequences are not allowed\n"
                    raise ValueError(err)

                sequence_list.append(seq)

        else:

            sequence = sequence.strip()
            if len(sequence) == 0:
                err = "\nempty sequences are not allowed\n"
                raise ValueError(err)

            return_singleton = True
            sequence_list.append(sequence)
    else:
        err = "\nsequence must either be a string or list of strings, one for each sequence\n\n"
        raise ValueError(err)

    # -------------------------------------------------------------------------
    # Deal with hitlist_size, e_value_cutoff

    hitlist_size = check.check_int(hitlist_size,
                                               "hitlist_size",
                                               minimum_allowed=1)
    e_value_cutoff = check.check_float(e_value_cutoff,
                                                   "e_value_cutoff",
                                                   minimum_allowed=0,
                                                   minimum_inclusive=False)

    # -------------------------------------------------------------------------
    # Deal with gapcosts

    try:
        g0 = int(gapcosts[0])
        g1 = int(gapcosts[1])
        if g0 < 1 or g1 < 1:
            raise ValueError
        if len(gapcosts) > 2:
            raise ValueError

        gapcosts = (g0,g1)
    except (IndexError,ValueError,TypeError,KeyError):
        err = "\ngapcosts should be a list of two integers greater than one\n"
        raise ValueError(err)

    return sequence_list, hitlist_size, e_value_cutoff, gapcosts, return_singleton



def read_blast_xml(blast_record):
    """
    Read BLAST XML format and return as a pandas data frame.

    Parameters
    ----------
    blast_record : biopython.Blast record or str
        either biopython Blast record or string pointing to blast xml file

    Returns
    -------
    results : list
        list of dataframes, one for each query in blast_record
    """

    # If argument is a string, treat as xml file
    if type(blast_record) is str:

        f = open(blast_record, 'r')
        p = NCBIXML.parse(f)

        blast_record = []
        for r in p:
            blast_record.append(r)
        f.close()

    out_df = []
    for record in blast_record:

        # Prepare DataFrame fields.
        data = {'accession': [],
                'hit_def': [],
                'hit_id': [],
                'title': [],
                'length': [],
                'e_value': [],
                'sequence': [],
                'subject_start': [],
                'subject_end':[],
                'query_start':[],
                'query_end':[],
                'query':[]}

        # Get alignments from blast result.
        for i, s in enumerate(record.alignments):
            data['accession'].append(s.accession)
            data['hit_def'].append(s.hit_def)
            data['hit_id'].append(s.hit_id)
            data['title'].append(s.title)
            data['length'].append(s.length)
            data['e_value'].append(s.hsps[0].expect)
            data['sequence'].append(s.hsps[0].sbjct)
            data['subject_start'].append(s.hsps[0].sbjct_start)
            data['subject_end'].append(s.hsps[0].sbjct_end)
            data['query_start'].append(s.hsps[0].query_start)
            data['query_end'].append(s.hsps[0].query_end)
            data['query'].append(record.query)

        if len(record.alignments) == 0:
            data['accession'].append(pd.NA)
            data['hit_def'].append(pd.NA)
            data['hit_id'].append(pd.NA)
            data['title'].append(pd.NA)
            data['length'].append(pd.NA)
            data['e_value'].append(pd.NA)
            data['sequence'].append(pd.NA)
            data['subject_start'].append(pd.NA)
            data['subject_end'].append(pd.NA)
            data['query_start'].append(pd.NA)
            data['query_end'].append(pd.NA)
            data['query'].append(record.query)

        # Port to DataFrame.
        out_df.append(pd.DataFrame(data))

    out_df = pd.concat(out_df)

    return out_df



"""
Construct a topiary dataframe from NCBI blast xml files.
"""

import topiary
from topiary import check, _private, ncbi

import pandas as pd
import numpy as np

from Bio import SeqIO

import os, glob, io

def df_from_blast_xml(xml_input):
    """
    Take a list of blast xml files and load in all sequences as a single
    topiary data frame. Parse meta data in an intelligent way, download
    sequences via entrez, and find unique taxonomic identifiers on the open
    tree of life.

    Parameters
    ----------
    xml_input : str or list
        blast xml files to load. This can have a few formats: 1) single xml file,
        2) list of xml files, or 3) directory. topiary will grab all files
        matching .xml in that directory.

    Returns
    -------
    topiary_dataframe : pandas.DataFrame
        topiary dataframe constructed by loading in all sequences from the
        xml file(s)
    """

    if type(xml_input) is str:

        # Looks like a file; treat as one
        if os.path.isfile(xml_input):
            xml_files = [xml_input]
        else:
            if os.path.isdir(xml_input):
                xml_files = glob.glob(os.path.join(xml_input,"*.xml"))
                xml_files.sort()
            else:
                xml_files = []

        if len(xml_files) == 0:
            err = f"\nCould not parse xml_input. Tried to read xml_input\n"
            err += f"'{xml_input}' as a file, then as a directory with .xml\n"
            err += "files. No xml files found. xml_input should be an xml file,\n"
            err += "list of xml files, or directory containing .xml files.\n\n"
            raise ValueError(err)

    else:
        xml_files = []
        if hasattr(xml_input,"__iter__") and type(xml_input) is not type:
            for x in xml_input:
                if os.path.isfile(x):
                    xml_files.append(x)
                else:
                    err = f"\nxml file '{x}' not found\n"
                    raise ValueError(err)
        else:
            err = f"\nCould not parse xml_input '{xml_input}'. Should be an xml\n"
            err += "file, list of xml files, or directory containing .xml files.\n\n"
            raise ValueError(err)


    # List to hold all hits and accession numbers to download
    all_hits = []
    to_download = []

    # For each xml file
    for i, xml in enumerate(xml_files):

        # Read xml
        tmp_df = ncbi.read_blast_xml(xml)

        # Go through and parse meta data
        for j in range(len(tmp_df)):

            accession = tmp_df.loc[j,"accession"]
            title = tmp_df.loc[j,"title"]
            evalue = tmp_df.loc[j,"e_value"]

            start = tmp_df.loc[j,"subject_start"]
            end = tmp_df.loc[j,"subject_end"]
            hit_info = ncbi.parse_ncbi_line(title,
                                            accession=accession)
            if hit_info is None:
                continue

            all_hits.append((xml,start,end,evalue,hit_info))
            to_download.append(hit_info["accession"])


    # Reduce to a unique set of accessions
    if len(set(to_download)) != len(to_download):
        unique_hits = []
        unique_download = []
        for i, d in enumerate(to_download):
            if d in unique_download:
                continue

            unique_hits.append(all_hits[i])
            unique_download.append(d)

        all_hits = unique_hits
        to_download = unique_download

    # Download sequences from entrez
    all_output = ncbi.get_sequences(to_download)

    # Capture sequences from the downloaded data
    captured = []
    for record in SeqIO.parse(io.StringIO(all_output), "fasta"):
        seq_id = str(record.id)
        sequence = str(record.seq)
        captured.append((seq_id,sequence))

    # Create a dictionary with appropriate keys to load into dataframe.
    key_list = ["name","species","sequence","uid","keep"]
    out = dict([(k,[]) for k in key_list])

    # Add blast-specific columns
    blast_key_list = ["accession","xml","length","evalue","start","end",
                      "structure","low_quality","precursor","predicted",
                      "isoform","hypothetical","partial","raw_line"]
    for k in blast_key_list:
        out[k] = []

    # Go through every hit
    for i in range(len(all_hits)):

        # Get information from previous few rounds
        seq = captured[i][1]
        accession = captured[i][0]

        xml = all_hits[i][0]
        start = all_hits[i][1]
        end = all_hits[i][2]
        evalue = all_hits[i][3]
        hit_info = all_hits[i][4]

        # Get hit_info, if k is in key_list
        for k in hit_info.keys():
            try:
                out[k].append(hit_info[k])
            except KeyError:
                pass

        # Overwrite accession from hit_info
        out["accession"][-1] = accession

        # Load info from blast itself
        out["xml"].append(xml)
        out["sequence"].append(seq)
        out["length"].append(len(seq))
        out["evalue"].append(evalue)
        out["start"].append(start-1)
        out["end"].append(end)

        out["uid"].append(_private.generate_uid())

        out["keep"].append(True)

    df = pd.DataFrame(out)
    df = check.check_topiary_dataframe(df)

    # Get ott id for all sequences, setting False for those that can't be
    # found/resolved
    df = topiary.get_ott_id(df)

    return df



"""
Construct a blast database.
"""

import topiary
from topiary import check

import os, subprocess, gzip, random, string

def make_blast_db(input_files,db_name,overwrite=False,makeblastdb_binary="makeblastdb"):
    """
    Make a protein blast database from a set of fasta input files.

    Parameters
    ----------
    input_files : list
        list of input files. Must be fasta files (with extension .faa) or
        zipped fasta files (with extension .faa.gz).
    db_name : str
        name of final database
    overwrite : bool, default=False
        whether or not to overwrite an existing database
    makeblastdb_binary : str, default="makeblastdb"
        makeblastdb binary to use

    Returns
    -------
    None
    """

    print("\nCreating blast database.\n",flush=True)

    # If the user passes in a string for input_files, make it into a list
    if type(input_files) is str:
        input_files = [input_files]

    if not hasattr(input_files,"__iter__") or type(input_files) is type:
        err = "\ninput_files should be a list of input files to put into the\n"
        err += "blast database.\n\n"
        raise ValueError(err)

    # Check for sanity of db_name
    db_name = str(db_name)

    # Process overwrite argument
    overwrite = check.check_bool(overwrite)

    # If we are overwriting an existing blast database.
    if os.path.exists(f"{db_name}.psq"):
        if overwrite:
            expected_extensions = ["pdb","pin","pot","ptf","phr","pjs","psq","pto"]
            for e in expected_extensions:
                try:
                    os.remove(f"{db_name}.{e}")
                except FileNotFoundError:
                    pass
        else:
            err = f"\nA blast database with name {db_name} already exists.\n\n"
            raise FileExistsError(err)

    # Make sure that makeblastdb is in the path
    try:
        subprocess.run([makeblastdb_binary])
    except FileNotFoundError:
        err = f"\makeblastdb_binary binary '{makeblastdb_binary}' not found in path\n\n"
        raise ValueError(err)

    rand_string = "".join([random.choice(string.ascii_letters) for _ in range(10)])
    tmp_faa_name = f"topiary_makeblastdb_{rand_string}.faa"

    # Combine files into a single temporary file
    out = open(tmp_faa_name,'a')
    for i, file in enumerate(input_files):
        print(f"Reading {file}",flush=True)

        if file[-7:] == ".faa.gz":
            with gzip.open(str(file)) as f:
                for line in f:
                    out.write(line.decode())
        elif file[-4:] == ".faa":
            with open(str(file)) as f:
                for line in f:
                    out.write(line)
        else:
            out.close()
            os.remove(tmp_faa_name)
            err = "\nInput files must have the extension .faa or .faa.gz.\n\n"
            raise ValueError(err)

    out.close()



    # Simple makeblastdb call
    cmd = [makeblastdb_binary,"-dbtype","prot","-in",tmp_faa_name,"-out",db_name]

    # This relatively complicated code is effectively subprocess.run, but
    # captures and dumps output to stdout as it is generated. This is useful
    # for a jupyter notebook, as the output appears in the notebook rather than
    # on the command line.
    popen = subprocess.Popen(cmd,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.STDOUT,
                             universal_newlines=True)
    for line in popen.stdout:
        print(line,end="",flush=True)
    popen.stdout.close()
    return_code = popen.wait()

    os.remove(tmp_faa_name)

    print("Done.",flush=True)



"""
Wrapper for generax to perform gene/species tree reconcilation.
"""

# raxml binary to use it not specified by user
GENERAX_BINARY = "generax"

import topiary
import topiary.external._interface as interface

import numpy as np

import subprocess, os, sys, time, random, string, shutil, copy
import multiprocessing as mp

def setup_generax(df,gene_tree,model,out_dir):
    """
    Setup a generax run directory.

    Parameters
    ----------
    df : pandas.DataFrame
        topiary data frame
    gene_tree : ete3.Tree or dendropy.Tree or str
        gene_tree with uid as taxon names. can be ete3 tree, dendropy tree, or
        newick file.
    model : str
        phylogenetic model to use (should match model used to generate gene_tree)
    out_dir : str
        output directory

    Returns
    -------
    None
    """

    # -------------------------------------------------------------------------
    # Create generax data structures

    # Load gene tree
    gene_tree = topiary.io.read_tree(gene_tree)

    # Arbitrarily resolve any polytomies in gene tree
    gene_tree.resolve_polytomy()

    # Map uid to ott in the dataframe
    uid_to_ott = {}
    for i in range(len(df)):
        uid = df.uid.iloc[i]
        ott = df.ott.iloc[i]
        uid_to_ott[uid] = ott

    # For generating mapping file, record which uid are associated with
    # which ott
    uid_in_gene_tree = []
    link_dict = {}
    for l in gene_tree.get_leaves():

        uid = l.name
        ott = uid_to_ott[uid]

        try:
            link_dict[ott].append(uid)
        except KeyError:
            link_dict[ott] = [uid]

        # Record that we saw this uid
        uid_in_gene_tree.append(uid)

    # Make df only have uid seen (will automatically trim down to only ott
    # of interest)
    mask = np.array([u in uid_in_gene_tree for u in df.uid],dtype=np.bool)
    df = df.loc[mask]

    # Get species tree corresponding to uid seen
    species_tree = topiary.get_species_tree(df)

    # Resolve polytomies and make sure all branch lenghts/supports have values
    species_tree.resolve_polytomy()

    # Go across tree and set taxon names to ott, supports to 1, and branch
    # lengths to 1.
    for n in species_tree.traverse():
        if n.is_leaf():
            n.name = copy.deepcopy(n.ott[0])
        if n.dist != 1:
            n.dist = 1
        if n.support != 1:
            n.support = 1

    # -------------------------------------------------------------------------
    # Write out generax input

    os.mkdir("working")

    # Construct the control file for generax
    control_out = []
    control_out.append("[FAMILIES]")
    control_out.append("- reconcile")
    control_out.append("starting_gene_tree = gene_tree.newick")
    control_out.append("alignment = alignment.phy")
    control_out.append("mapping = mapping.link")
    control_out.append(f"subst_model = {model}")

    # Write out control file
    f = open(os.path.join(out_dir,"control.txt"),"w")
    f.write("\n".join(control_out))
    f.close()

    # Write out .phy file
    topiary.write_phy(df,os.path.join(out_dir,"alignment.phy"),
                      seq_column="alignment")

    # Write out gene tree
    gene_tree.write(outfile=os.path.join(out_dir,"gene_tree.newick"),
                    format=5)

    # Write out species tree
    species_tree.write(outfile=os.path.join(out_dir,"species_tree.newick"),
                       format=5)

    # Write out link file
    f = open(os.path.join(out_dir,"mapping.link"),"w")
    for k in link_dict:
        f.write(f"{k}:")
        f.write(";".join(link_dict[k]))
        f.write("\n")
    f.close()


def run_generax(run_directory,
                allow_horizontal_transfer=False,
                seed=None,
                generax_binary=GENERAX_BINARY,
                log_to_stdout=True,
                other_args=[]):

    """
    Run generax. Creates a working directory, copies in the relevant files, runs
    there, and then returns to the previous directory.

    Parameters
    ----------
    run_directory : str
        directory in which to do calculation
    allow_horizontal_transfer : bool, default=False
        whether or not to allow horizontal gene transfer. This corresponds to
        the UndatedDTL (horizontal) vs UndatedDL (no horizontal) models
    seed : bool or int or str, optional
        If true, pass a randomly generated seed to generax. If int or str, use
        that as the seed (passed via --seed).
    generax_binary : str, optional
        generax binary to use
    log_to_stdout : bool, default=True
        capture log and write to std out.
    other_args : list, optional
        other arguments to pass to generax

    Returns
    -------
    generax_cmd : str
        string representation of command passed to generax
    """

    cmd = ["generax"]
    cmd.extend(["--families","control.txt"])
    cmd.extend(["--species-tree","species_tree.newick"])
    cmd.extend(["--prefix","result"])

    if allow_horizontal_transfer:
        model = cmd.extend(["--rec-model","UndatedDTL"])
    else:
        model = cmd.extend(["--rec-model","UndatedDL"])

    # seed argument is overloaded. Interpret based on type
    if seed is not None:
        if type(seed) is int:
            cmd.extend(["--seed",f"{seed:d}"])
        elif type(seed) is str:

            try:
                int(seed)
            except ValueError:
                err = f"seed {seed} could not be interpreted as an int\n"
                raise ValueError(err)

            cmd.extend(["--seed",seed])
        elif type(seed) is bool:
            if seed:
                cmd.extend(["--seed",interface.gen_seed()])
        else:
            err = "seed must be True/False, int, or string representation of int\n"
            raise ValueError(err)

    # Grab other args
    cmd.extend(other_args)

    # Make sure that generax is in the path
    try:
        subprocess.run([generax_binary])
    except FileNotFoundError:
        err = f"\ngenerax binary '{generax_binary}' not found in path\n\n"
        raise ValueError(err)

    if not os.path.exists(run_directory):
        err = f"\nrun_directory '{run_directory}' not found.\n\n"
        raise FileNotFoundError(err)

    if not os.path.isdir(run_directory):
        err = f"\nrun_directory must be a directory not found.\n\n"
        raise ValueError(err)

    required_files = ["control.txt","alignment.phy",
                      "gene_tree.newick","species_tree.newick",
                      "mapping.link"]
    for f in required_files:
        filename = os.path.join(run_directory,f)
        if not os.path.exists(filename):
            err = f"\nrun_directory '{run_directory}' does not have all required\n"
            err += "files. It should have the following files:\n"
            for r in required_files:
                err += f"    {r}\n"
            err += "\n"
            raise FileNotFoundError(err)

    log_file = None
    if log_to_stdout:
        log_file = os.path.join("result","generax.log")

    # Launch run
    interface.launch(cmd,run_directory,log_file)

    return " ".join(cmd)



"""
Load a seed data frame and extract information needed to set up the
ancestral sequence reconstruction calculation.
"""

import topiary

import numpy as np
import pandas as pd

import re, itertools
from string import ascii_lowercase, digits

def _get_string_variants(some_string,spacers):
    """
    Get regular expressions describing strings with variable spacers. For
    example LY96, LY-96, LY 96, LY_96. This function looks for elements in
    spacers or transitions between letters and numbers and builds regular
    expressions to match those patterns. It makes all letters lowercase
    and thus expects to be used in a regular expression ignoring case.

    Examples (for spacers = [" ","-","_","."]):
        "LY96" -> "ly[\ \-_\.]*96"
        "MD-2" -> "md[\ \-_\.]*2"
        "myeloid factor 2" -> "myeloid[\ \-_\.]*factor[\ \-_\.]*2"
        "CDR2L" -> "cdr[\ \-_\.]*2[\ \-_\.]*L"

    Parameters
    ----------
    some_string : string to process
        spacers: list of characters to recognize as spacers

    Return
    ------
        regular expression as a string
    """

    # Regular expression describing spacers
    spacer_re = "".join([re.escape(s) for s in spacers])
    spacer_re = f"[{spacer_re}]*"

    # Make lowercase and strip leading/trailing white space
    lower = some_string.lower().strip()

    # Get digit, letter, and spacer characters
    is_digit = np.zeros(len(lower),dtype=bool)
    is_letter = np.zeros(len(lower),dtype=bool)
    is_spacer = np.zeros(len(lower),dtype=bool)
    for i, s in enumerate(lower):
        if s in digits:
            is_digit[i] = True
        elif s in ascii_lowercase:
            is_letter[i] = True
        elif s in spacers:
            is_spacer[i] = True
        else:
            pass

    # Go through sequence and build pattern character by character
    pattern = []
    for i in range(len(lower)-1):

        # If we hit a spacer, put in non-greedy match
        if is_spacer[i]:
            pattern.append(spacer_re)
            continue

        # Append the letter/digit we saw
        pattern.append(lower[i])

        # Digit followed by letter
        if is_digit[i] and is_letter[i+1]:
            pattern.append(spacer_re)

        # Letter followed by digit
        if is_letter[i] and is_digit[i+1]:
            pattern.append(spacer_re)

    # Last letter
    pattern.append(lower[-1])

    return "".join(pattern)


def _get_alias_regex(list_of_names,spacers=[" ","-","_","."]):
    """
    Produce a regular expression that matches all patterns in list of names,
    expanding all spacers characters to look for all possiblities at that
    site. For example, names might be ["LY96", "MD2", "ESOP1"]. These will be
    put into _get_string_variants, yielding a final regular expression:
        "ly[\ \-_.]*96|md[\ \-_.]*2|esop[\ \-_.]*1"

    Parameters
    ----------
        list_of_names: list of strings from which to build the regex.
        spacers: list of characters to recognize as spacers

    Return
    ------
        compiled regular expression instance
    """

    # Make unique set of names
    list_of_names = list(set(list_of_names))



    # Make regular expression for each name
    all_names = []
    for n in list_of_names:
        all_names.append(_get_string_variants(n,spacers))

    # Unique set of regex
    all_names = list(set(all_names))

    # sort for testing purposes so regex has predictable order
    all_names.sort()

    return re.compile("|".join(all_names),flags=re.IGNORECASE)


def load_seed_dataframe(df):
    """
    Load a seed data frame and extract information needed to set up the
    ancestral sequence reconstruction calculation.

    Parameters
    ----------
    df: pandas.DataFrame or str
        seed dataframe containing seed sequences to launch the analysis. df can
        be a pandas dataframe or a string pointing to a spreadsheet file.

    Returns
    -------
    topiary_dataframe : pandas.DataFrame
        new topiary dataframe build from the seed dataframe
    phylo_context : str
        string descriptor of the phylogenetic context on opentreeoflife.org
        (i.e. "Animals" or "All life")
    key_species : list
        list if key species to keep during the analysis
    paralog_patterns : list
        list of compiled regular expressions (extracted from aliases) to use to
        try to match paralogs.

    Notes
    -----

    The seed dataframe is expected to have at least four columns:

    + `species`: species names for seed sequences in binomial format (i.e.
      Homo sapiens or Mus musculus)
    + `name`: name of each sequence (i.e. LY96)
    + `aliases`: other names for this sequence found in different
      databases/species, separated by ; (i.e. LY96;MD2;ESOP1)
    + `sequence`: amino acid sequences for these proteins.

    Other columns in the dataframe are kept but not used by topiary.
    """

    # -----------------------------------------------------------------------
    # Load dataframe and check it's sanity

    if type(df) is not type(pd.DataFrame()):

        if type(df) is str:
            extension = df.split(".")[-1].strip().lower()
            if extension in ["xls","xlsx"]:
                df = pd.read_excel(df)
            elif extension == "csv":
                df = pd.read_csv(df,sep=",")
            elif extension == "tsv":
                df = pd.read_csv(df,sep="\t")
            else:
                df = pd.read_csv(df,sep=None,engine="python")
        else:
            err = f"Could not figure out how to read df '{df}'\n\n"
            raise ValueError(err)

    required_columns = ["species","name","aliases","sequence"]
    for c in required_columns:

        try:
            df[c]
        except KeyError:
            err = f"\nDataframe must have a {c} column. Required columns are:\n"
            for r in required_columns:
                err += f"    {r}\n"
            err += "\n"
            raise ValueError(err)

    # -----------------------------------------------------------------------
    # Get key_species and phylogenetic context

    key_species = list(np.unique(df.loc[:,"species"]))
    key_species.sort() # sort alphabetically

    phylo_context = topiary.opentree.get_phylo_context(key_species)

    # -----------------------------------------------------------------------
    # Construct paralog_patterns

    # Build a paralog_patterns dictionary
    alias_dict = {}
    for idx in df.index:

        # Construct list of all aliases given by the user for a given name
        key = str(df.loc[idx,"name"])
        values = [v.strip() for v in df.loc[idx,"aliases"].split(";")]

        try:
            alias_dict[key].extend(values)
        except KeyError:
            alias_dict[key] = values[:]

    paralog_patterns = {}
    for a in alias_dict:

        # Put name itself in as an alias
        alias_dict[a].append(a)

        # Get regular expression representations of these patterns
        paralog_patterns[a] = _get_alias_regex(alias_dict[a])

    # Load newly created paralog patterns into aliases column
    new_aliases = []
    for idx in df.index:

        name = df.loc[idx,"name"]
        new_aliases.append(paralog_patterns[name].pattern)

    df.loc[:,"aliases"] = new_aliases


    # -----------------------------------------------------------------------
    # Convert dataframe into a topiary dataframe

    # add always keep column to this dataframe
    df["always_keep"] = True

    # Drop on uid column to avoid warning when topiary creates in arg_processor
    try:
        df["uid"]
    except KeyError:
        df["uid"] = topiary._private.generate_uid(len(df["always_keep"]))

    df = topiary.check.check_topiary_dataframe(df)

    return df, phylo_context, key_species, paralog_patterns



"""
Interface to open tree of life database.
"""

from .util import is_allowed_phylo_context, get_phylo_context, species_to_ott
from .ott import get_ott_id
from .tree import get_species_tree

import pandas as pd
import os

# -----------------------------------------------------------------------------
# Load a csv file mapping between open tree phylogenetic context and ncbi
# taxid

def _load_phylo_to_taxid(csv):
    """
    Load the mapping between open tree of life phylogenetic contexts and NCBI
    taxids.

    Parameters
    ----------
        csv: csv file with mapping

    Return
    ------
        dictionary with mapping
    """

    # Read csv
    df = pd.read_csv(csv)

    # Go through dataframe
    phylo_to_taxid = {}
    for idx in df.index:

        # Cast context and taxid as strings
        context = str(df.loc[idx,"phylo_context"])
        taxid = str(df.loc[idx,"taxid"]).strip()

        # Pull out "None" taxid
        if taxid == "None":
            taxid = None
            phylo_to_taxid[context] = taxid
            continue

        # Pull out tuple taxid
        if taxid.startswith("("):
            chopped = taxid[1:-1].split(",")
            taxid = tuple([int(c) for c in chopped])
            phylo_to_taxid[context] = taxid
            continue

        # Grab simple integer taxid
        taxid = int(taxid)
        phylo_to_taxid[context] = taxid

    # Update keys to include uppercase and lowercase versions
    keys = list(phylo_to_taxid.keys())
    for k in keys:
        phylo_to_taxid[k.lower()] = phylo_to_taxid[k]
        phylo_to_taxid[k.upper()] = phylo_to_taxid[k]

    return phylo_to_taxid

# Get location of topiary in the file system
_location = os.path.dirname(os.path.realpath(__file__))
_context_to_taxid_file = os.path.join(_location,"phylo_context_to_taxid.csv")

# Try to read csv file with the ott context to taxid map
try:
    phylo_to_taxid = _load_phylo_to_taxid(_context_to_taxid_file)
except Exception as e:
    err = f"\nCould not load {_context_to_taxid_file}\n\n"
    raise Exception from e(err)



"""
Load a tree into an ete3 tree data structure.
"""

import ete3
from ete3 import Tree
import dendropy as dp

def read_tree(tree,fmt=None):
    """
    Load a tree into an ete3 tree data structure.

    Parameters
    ----------
    tree : ete3.Tree or dendropy.Tree or str
        some sort of tree. can be an ete3.Tree (returns self), a dendropy Tree
        (converts to newick and drops root), a newick file or a newick string.
    fmt : int or None
        format for reading tree from newick. 0-9 or 100. (See Notes for what
        these mean). If fmt is None, try to parse without a format descriptor,
        then these formats in numerical order.

    Returns
    -------
    tree : ete3.Tree
        an ete3 tree object.

    Notes
    -----
    `fmt` number is read directly by ete3. See their documentation for how these
    are read (http://etetoolkit.org/docs/latest/tutorial/tutorial_trees.html#reading-and-writing-newick-trees).
    As of ETE3.1.1, these numbers mean:

    + 0: flexible with support values
    + 1: flexible with internal node names
    + 2: all branches + leaf names + internal supports
    + 3: all branches + all names
    + 4: leaf branches + leaf names
    + 5: internal and leaf branches + leaf names
    + 6: internal branches + leaf names
    + 7: leaf branches + all names
    + 8: all names
    + 9: leaf names
    + 100: topology only

    """

    # Already an ete3 tree.
    if type(tree) is ete3.TreeNode:
        return tree

    # Convert dendropy tree into newick (drop root)
    if type(tree) is dp.Tree:
        tree = tree.as_string(schema="newick",suppress_rooting=True)

    # If we get here, we need to convert. If fmt is not specified, try to parse
    # without a format string.
    if fmt is None:


        try:
            t = Tree(tree)
        except ete3.parser.newick.NewickError:

            # Try all possible formats now, in succession
            w = "\n\nCould not parse tree without format string. Going to try different\n"
            w += "formats. Please check output carefully.\n\n"
            print(w)

            formats = list(range(10))
            formats.append(100)

            t = None
            for f in formats:
                try:
                    t = Tree(tree,format=f)
                    w = f"\n\nSuccessfully parsed tree with format style {f}.\n"
                    w += "Please see ete3 documentation for details:\n\n"
                    w += "http://etetoolkit.org/docs/latest/tutorial/tutorial_trees.html#reading-and-writing-newick-trees\n\n"
                    print(w)
                    break

                except ete3.parser.newick.NewickError:
                    continue

            if t is None:
                err = "\n\nCould not parse tree!\n\n"
                raise ValueError(err)

    else:
        # Try a conversion with the specified format
        t = Tree(tree,format=fmt)

    return t
